/*
 * Copyright 1994, 1995 Vectaport Inc., Cartoactive Systems
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Map/mapelt.h>

#include <iostream.h>
#include <stdio.h>
#include <string.h>

/*****************************************************************************/

MapElement::MapElement(int eid, MapFeatureType t, int n, double* x, double* y,
		       void* g) 
: MapFeature(nil, eid, t) {
    init_points(n, x, y);
    _gr = g;
    _textstr = nil;
}

MapElement::MapElement(int eid, char* str, void* g) : MapFeature(nil, eid, MapFeature::TextType) {
    _en = 0;
    _ex = _ey = nil;
    _gr = g;
    _textstr = strdup(str);
}

MapElement::MapElement() : MapFeature(nil, 0, MapFeature::PointType) {
    _en = 0;
    _gr = _ex = _ey = nil;
}

MapElement::~MapElement() {
    delete [] _ex;
    delete [] _ey;
    delete [] _textstr;
}

void MapElement::init_points(int n, double* x, double* y) {
    _en = n;
    _ex = new double[_en];
    _ey = new double[_en];
    memcpy(_ex, x, _en * sizeof(double));
    memcpy(_ey, y, _en * sizeof(double));
}

void* MapElement::graphic() const {
    return _gr;
}

int MapElement::points(const double*& x, const double*& y) const {
    x = _ex;
    y = _ey;
    return _en;
}

void MapElement::points(double*& x, double*&y, int& n) const {
    n = _en;
    x = new double[n];
    y = new double[n];
    memcpy(x, _ex, n * sizeof(double));
    memcpy(y, _ey, n * sizeof(double));
}

#define EPS10 1.0e-10

void MapElement::points(PJ* proj, double*& x, double*&y, int& n) const {
    n = _en;
    x = new double[n];
    y = new double[n];
    for (int i = 0; i < _en; i++) {
	UV data;
	data.u = _ex[i] * DEG_TO_RAD;
	data.v = _ey[i] * DEG_TO_RAD;
	data = pj_fwd(data, proj);
	if (data.u == HUGE_VAL) {
	    /* cerr << "projection error\n"; */
	    data.u = _ex[i] * DEG_TO_RAD;
	    data.v = _ey[i] * DEG_TO_RAD;
	    if (_ex[i]==90.0)
		data.u -= 2 * EPS10;
	    if (_ex[i]==-90.0)
		data.u += 2 * EPS10;
	    if (_ey[i]==90.0)
		data.v -= 2 * EPS10;
	    if (_ey[i]==-90.0)
		data.v += 2 * EPS10;
	    data = pj_fwd(data, proj);
	    if (data.u == HUGE_VAL)
		cerr << "projection error again\n";
	}
	x[i] = data.u;
	y[i] = data.v;
    }
}

int MapElement::num_points() const {
    return _en;
}

void MapElement::point(int pn, double& x, double& y) const {
    x = _ex[pn];
    y = _ey[pn];
}

const char* MapElement::text() const {
    return _textstr;
}

void MapElement::graphic(void* g) {
    _gr = g;
}

void MapElement::points(int n, double* x, double* y) {
    init_points(n,x,y);
}

void MapElement::text(char* s) {
  _textstr = strdup(s);
}

const char* MapElement::describe() const {
  return nil;
}

char GeoPoint::_desc[256];

GeoPoint::GeoPoint(int id, double x, double y, void* g) 
: MapElement(id, PointType, 1, &x, &y, g) {
}

const char* GeoPoint::describe() const {
  sprintf(_desc, "Geographic Point at %0.2fN %0.2fE", _ey[0], _ex[0]);
  return _desc;
}
